import requests
import pickle
import pandas as pd
import json
import time 
import re 
import importlib
from Ambra_Clone.V2.ambra_environment import ambra_environment
import __main__ as main
if not hasattr(main, '__file__'):
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm
"""
Template CSV runner -- example here is for a study/sync api call. study_id parameter uses the uuid column of the csv file specified by the user
User input can be caputured by setting a variable = input("prompt: ")
"""
class runner:
    def __init__(self,url,sid,endpoint,reader_data:pd.DataFrame,params={"uuid":"{uuid}"},delimiter=","):
        self.url = url
        self.sid = sid
        self.endpoint = endpoint
        # self.csv_file = csv_file
        self.params = params
        self.failed_df = pd.DataFrame(columns=reader_data.columns)
        self.reader_data = reader_data
        self.failed_df = pd.DataFrame(columns=self.reader_data.columns)
        self.failed_df['reason'] = ""

        self.responses = []
        self.env = ambra_environment(sid,url)
        self.prep_requests()

    def prep_requests(self,df=None):
        self.reqs = []
        if df == None:
            df = self.reader_data
        pbar = tqdm(df.iterrows(),total=len(df))
        pbar.set_description("loading file")
        for i, csv_row in pbar:
            if self.endpoint == "/bundle": 
                bundle_reqs = []
                if isinstance(self.params,list):
                    for req in self.params:
                        r = {}
                        for k,v in req.items():
                            if isinstance(v,str):
                                r[k] = self.make_replacements(v,csv_row,re.findall('((?<!\{)\{(?!\{){1}[^}]+\})',v))
                            else:
                                r[k] = v
                        r['sid'] = self.sid
                        bundle_reqs.append(r)
                    self.reqs.append(json.dumps(bundle_reqs))
                    # response = self.send_requests_bundle(reqs)
                    # self.responses.append(response)
                else:
                    print("params must be a list of dictionaries when using the bundle endpoint")
                    raise "params must be a list of dictionaries when using the bundle endpoint"
            else:
                req = self.create_request(csv_row)
                self.reqs.append(req)
    def preview_data(self):
        pd.DataFrame(self.reqs).head()

    def run(self):
        self.success_count = 0
        self.failed_count = 0

        self.responses = self.env.multiprocess_ambra_request(self.reqs,endpoint=self.endpoint,pbar=True)
        self._handle_responses(self.reqs)
        # self.failed_df.to_csv("failed.csv")
        #print the number of failed studies:
        print("failed_studies ", self.failed_count)
        #number of successful studies: 
        print("successful_studies ", self.success_count)

    def run_sample(self):
        sample_data = self.reqs[:5]
        response = []
        for row in sample_data:
            print("request: ",row)
            response = self.env.handle_ambra_request(self.endpoint,row)
            print("response: ", response,"\n")
                
    def make_replacements(self,v,csv_row,tokens_to_replace):
        for token in tokens_to_replace:
            csv_fieldname = token.replace("{","").replace("}","")
            if csv_fieldname in csv_row:
                replace_value = csv_row[csv_fieldname]
                v =  v.replace(token,replace_value)
            else:
                print(token + " not found in csv")
        return v
    def create_request(self,csv_row):
        req = {
                'URL':self.endpoint,
                "sid":self.sid,
            }
        for k in self.params:
            replace_value = None
            v = self.params[k]
            tokens_to_replace = re.findall('((?<!\{)\{(?!\{){1}[^}]+\})',v)
            if len(tokens_to_replace) == 0:
                req[k] = v
            else:
                req[k] = self.make_replacements(v,csv_row,tokens_to_replace)
        return req

    def _handle_responses(self,reqs,response=None):
        if response == None:
            response = self.responses
        for ind, r in enumerate(response):
            if len(r)>2 and r[0]=="{" and r[-1]=="}":
                r = json.loads(r)
                # sort dictionary by key
                r = {k: r[k] for k in sorted(r)}
            if isinstance(r,list):
                for res in r: 
                    if res.get('status') != "OK":             
                        failed_study = self.reader_data.iloc[[ind]].to_dict('records')[0]
                        failed_study['reason'] = str(res)
                        self.failed_count+=1
                        self.failed_df = self.failed_df.append(failed_study,ignore_index=True)
                        break
                    self.success_count+=1
            elif r.get('status') != "OK":
                self.failed_count+=1
                failed_study = self.reader_data.iloc[[ind]].copy()
                failed_study['reason'] = str(r)
                self.failed_df = self.failed_df.append(failed_study,ignore_index=True)
            else:
                self.success_count+=1
        return self.responses
    
    def summarize_responses(self):
        #print a success and failure count. Failures are counted from the failed_df dataframe. Su
        print("successes: ", self.success_count)
        print("failures: ", self.failed_count)
        print("distinct errors: ")
        #print distinct values in self.failed_df['reason'] 
        print(self.failed_df['reason'].unique())

    def output_responses_to_csv(self,filename):
        pd.DataFrame(self.responses).to_csv(filename)
    def output_failed_to_csv(self,filename):
        self.failed_df.to_csv(filename)
    def rerun_failed(self):
        self.prep_requests(self.failed_df)
        self.run()
    def pickle_me(self,filepath):
        #pickle self
        if filepath[-4:] != ".pkl":
            filepath = filepath+".pkl"
        with open(filepath,"wb") as f:
            pickle.dump(self,f)
    


from .ambra_environment import ambra_environment

import requests
from json import loads,dump,load,dumps
import os
from tkinter.filedialog import asksaveasfilename, askopenfilename
from tkinter import Tk
from csv import DictWriter
import __main__ as main
from csv import DictWriter
if not hasattr(main, '__file__'):
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

class account_configuration:
    def __init__(self, ambra_environment=None,account_id:str=None,name="",clone_items:list=[]):
        self.init_config = {
        "skip_fields": {
            "account": ["name","role_name","user_account_id","vanity","saml_redirect_url","md5"],
            "customfields":["customfield_mapping"],
            "route": ["namespace_name","actions:name","actions:destination"],
            "group": ["anonymize_at_ingress","namespace_id"],
            "location":["anonymize_at_ingress","namespace_id"],
            "dictionary":["attachments:uuid"],
            "role":["permissions:viewer3_config"],
            "terminology":["vanity"],
            "account/user":["user_account_id"],
            "namespace/settings":["linked_namespace_name","location_id","group_id","status"],
            "namespace/anonymize":["dicom_tag_map"],
            "webhook":["last_error"]
            },
            "special_fields":{},
            "error_dict": {
                "DUP_SHARE_CODE":"share_code",
                "INVALID_OTHER_NAMESPACES": "other_namespaces"
                }, 
            "special_uuid": {},
            "skip_uuid": [],
            "id_map" : {},
            "never_skip_fields":["filter_field"],
            "available_api_strs": [
                    "account",
                    "customfield",
                    "customfield/mapping",
                    "role",
                    "group",
                    "location",
                    "account/user",
                    "webhook",
                    "route",
                    "group/user",
                    "location/user",
                    "namespace/settings",
                    "namespace/anonymize",
                    "namespace/study/defaults",
                    "namespace/event/defaults",
                    "terminology",
                    "hl7/template",
                    "hl7/transform",
                    "dictionary",
                    "mail/template",
                    "radreport/template",
                    "account/can/share",
                    "site",
                    "destination",
                    "node"],
            "list_api_strs": {"customfield":"customfields",
                "role":"roles",
                "group":"groups",
                "location":"locations",
                "webhook":"webhooks",
                "route":"routes",
                "hl7/template":"templates",
                "hl7/transform":"transforms",
                "dictionary":"dictionaries",
                "site":"sites",
                "radreport/template":"templates",
                "mail/template":"templates",
                "account/user":"users",
                "group/user":"users",
                "location/user":"users",
                "terminology":"tags",
                "account/can/share":"rules",
                "destination":"destinations",
                "node":"nodes"},
            "required_add_fields": {
                "customfield":["object","type"],
                "customfield/mapping":["from_customfield_id",'to_customfield_id'],
                "role":[],
                "group":[],
                "location":[],
                "account/user":["email","role_id"],
                "webhook":["url","method","event","cron"],
                "route":["conditions","actions","namespace_id"],
                "group/user":["user_id","role_id"],
                "location/user":["user_id","role_id"],
                "hl7/template":["body"],
                "hl7/transform":["conditions","replacements","order_by"],
                "dictionary":["object","lookup","replace","case_sensitive"],
                "mail/template":["type"],
                "radreport/template":["type","body"],
                "site":[],
                "destination":["node_id","aetitle","address","port"],
                "node":["type"]},
                "cast_as_string":["rules","conditions","dicom_tag_map"]
                }
        for config in self.init_config:
            setattr(self,config,self.init_config[config])
        self.config_dirname = os.path.dirname(__file__)
        filename = os.path.join(self.config_dirname, "init_config.json")
        self.namespace_properies_by_id = {}
        
        if ambra_environment==None: 
            self.load_from_file()
        else:
            self.account = {}
            self.env = ambra_environment
            if account_id in ["",None] and name != '': 
                try:
                    account_response = self.env.handle_ambra_request("/account/add",data={'sid':self.env.sid,"name":name},throw=1)
                    self.account = account_response
                    self.account_id = account_response['uuid']
                    self.account['name'] = name
                except Exception as err:
                    raise Exception("failed adding account " + name + " - " + str(err))
            else:
                self.account_id=account_id
                self.account = self.env.handle_ambra_request("account/get",data={"uuid":account_id},throw=1)
            self.permissions = self.env.handle_ambra_request("/session/permissions",data={"sid":self.env.sid,"account_id":account_id})
            self.reload_account_configuration(clone_items)
        
        
    def reload_account_configuration(self,api_strs:list=None):
        print("Loading account configuration for "+self.account.get("name")+" - uuid: "+self.account.get("uuid"))
        #api calls that have a /list endpoint
        # if api_strs in [None,"all",['all']] :
        api_strs = self.available_api_strs
        # else:
        #     api_strs = api_strs
        pbar = tqdm(api_strs,total=len(api_strs))
        for apistr in pbar:
            if apistr == "account":
                try:
                    setattr(self,"account",self.env.handle_ambra_request("/account/get",data={"sid":self.env.sid,"uuid":self.account_id},throw=1))   
                except:
                    raise Exception("Unable to get account information for "+self.account_id+ " please ensure session information is correct \n URL:"+self.env.url+"\n SID:"+self.env.sid)
            else:
                try:
                    pbar.set_description("loading "+apistr)
                    if apistr in self.list_api_strs.keys() and 'user' not in apistr:
                        data = {"sid":self.env.sid,"account_id":self.account_id}
                        tmpstr = self.list_api_strs[apistr]
                        response = self.env.handle_ambra_request("/{apistr}/list".format(url=self.env.url,apistr=apistr),data=data)
                        if response is None:
                            setattr(self,apistr,[])
                        else:
                            if apistr == "customfield":
                                setattr(self,"customfield/mapping",[])
                                for cf in response['customfields']:
                                    setattr(self,"customfield/mapping",getattr(self,"customfield/mapping")+cf['customfield_mapping'])
                            setattr(self,apistr,response[tmpstr])
                    
                    elif apistr == "account/user":
                        # try:
                            out = self.env.handle_ambra_request(apistr+"/list",data={"uuid":self.account_id})
                            out = out.get('users')
                            for user in out: 
                                user['linked_namespace_name'] = self.account.get('name')
                                user['namespace_id'] = self.account.get('namespace_id')
                                user['account_id'] = self.account.get('uuid')
                            setattr(self,apistr,out)
                        # except:
                        #     print("skipping account/user")
                            
                    elif 'namespace' in apistr or apistr in ["group/user","location/user"]:
                        threaded_requests = []
                        if apistr in ["group/user","location/user"]:
                            endpnt = apistr+"/list"
                            field = "uuid"
                            out = {}
                            namespaces = []
                            for x in getattr(self,apistr.split("/")[0]):
                                threaded_requests.append({"uuid":x[field],"sid":self.env.sid})
                                namespaces.append(x)
                            threaded_response = self.env.multiprocess_ambra_request(threaded_requests,endpnt)
                        elif 'namespace' in apistr:
                            endpnt = apistr
                            field = "namespace_id"
                            out = []
                            namespaces = [self.account]+self.group+self.location
                            for ns in namespaces:
                                threaded_requests.append({"uuid":ns['namespace_id'],"sid":self.env.sid})
                            threaded_response = self.env.multiprocess_ambra_request(threaded_requests,endpnt)
                            
                        for i in range(len(threaded_response)):
                            res = threaded_response[i]
                            x = namespaces[i]
                            if 'user' in apistr:
                                out[x[field]] = res['users']
                                for user in out[x[field]]:
                                    user['linked_namespace_name'] = x['name']
                                    user['uuid'] = x['uuid']
                            elif 'namespace' in apistr:
                                
                                if 'rules' in res:
                                    rules = loads(res['rules'])  
                                    for key in list(rules): # this loop gets rid of the other_ingress_tags all other_ingress_tags from account settings are listed but cannot be set in namespace/anonymize. Expecting OTHER_INGRESS_TAGS : ANONYMIZATION VALUE instead
                                        if '(' in key  and ',' in key:
                                            # rules['OTHER_INGRESS_TAGS'] = rules[key]
                                            del rules[key]
                                    if len(rules) > 0:
                                        res['rules'] = dumps(rules)
                                    else:
                                        del res['rules']
                                res['linked_namespace_name'] = x['name']
                                res['uuid'] =x['namespace_id']
                                res['location_id'] = x['uuid']
                                out.append(res)
                        setattr(self,apistr,out)
                except:
                    print("skipping "+apistr)
        self.reload_namespace_properties()
    def reload_namespace_properties(self):
        self.namespace_properies_by_id[self.account['namespace_id']] = {"account_id":self.account['uuid'],"name":self.account['name']}
        for g in self.group:    
            self.namespace_properies_by_id[g['namespace_id']] = {"group_id":g['uuid'],"name":g['name']}
        for l in self.location:
            self.namespace_properies_by_id[l['namespace_id']] = {"location_id":l['uuid'],"name":l['name']}
        for anon in getattr(self,"namespace/anonymize"):
            self.namespace_properies_by_id[anon['uuid']].update({"namespace/anonymize":anon})
        for default in getattr(self,"namespace/settings"):
            self.namespace_properies_by_id[default['uuid']].update({"namespace/settings":default})
    def get_account_json(self):
        return {d: self.dict_sort(self.__dict__.get(d)) for d in self.__dict__ if d in self.available_api_strs}
    def generate_account_file(self,account_file=None):
        if account_file == None:
            root = Tk()
            
            account_file = asksaveasfilename(defaultextension='.json',initialfile=self.account['name']+".json")
            root.destroy()
        dump(self.get_account_json(),open(account_file,'w',newline=""),indent=4)
    def dict_sort(self,data):
        #data can be a list or a dictionary. Return a sorted version of the data
        #sort all sub dictionaries and lists
        #sort list of dictionaries by the dictionary 'name' or 'user_name' key
        if type(data) == list:
            for i in range(len(data)):
                data[i] = self.dict_sort(data[i])
            if len(data) > 0 and type(data[0]) == dict:
                if 'name' in data[0]:
                    return sorted(data,key=lambda x: x['name'])
                elif 'user_name' in data[0]:
                    return sorted(data,key=lambda x: x['user_name'])
                elif 'linked_namespace_name' in data[0]:
                    return sorted(data,key=lambda x: x['linked_namespace_name'])
                else:
                    return sorted(data,key=lambda x: x[sorted(list(x.keys()))[0]])
            else:
                return data
        elif type(data) == dict:
            for key in data:
                data[key] = self.dict_sort(data[key])
            return dict(sorted(data.items()))
        else:
            return data
            
        
    def load_from_file(self):
        root = Tk()
        account_file = askopenfilename(title="Select account configuration json file",filetypes=[("accont_configuration",".json")])
        root.destroy()
        config = load(open(account_file,'r'))
        for k,v in config.items(): 
            setattr(self,k,v)
        self.reload_namespace_properties()
    # clone_items= [
    # "customfield",
    # "role",
    # "group",
    # "location",
    # "account/users",
    # "webhook",
    # "route",
    # "group_location_users",
    # "namespace_settings",
    # "terminology",
    # "hl7/template",
    # "hl7/transform",
    # "dictionary",
    # "mail_templates",
    # "account_settings",
    # "radreport/template",
    # "share_settings",
    # "site",
    # "destinations",
    # "nodes"]
### FOR TESTING
# calyx_uat = ambra_environment("785e516b-0861-4d5b-88ab-f5510e76538a",url = "https://mi.uat.calyx.ai/api/v3")
# act_config = account_configuration(calyx_uat,"d9084e43-6c35-4dc5-a852-caea6e618623")
# ambra_uat = ambra_environment("da891b40-3d34-49b0-ad7a-c76ef8b4b8c7",url = "https://uat.ambrahealth.com/api/v3")
# act_config = account_configuration(ambra_uat,"57ef3334-f629-4422-9406-e2bf3a8441fc")
# act_config.generate_account_file()


# local_load = account_configuration()


import json
import requests
from time import sleep
from multiprocessing.pool import ThreadPool
from tqdm import tqdm
from functools import partial
import re
import __main__ as main
if not hasattr(main, '__file__'):
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

class ambra_environment:
    def __init__(self, sid:str,url="https://access.ambrahealth.com/api/v3"):
        """Authenticate ambra session to make api calls

        Args:
            sid (str): valid session id for cloning 
            url (str, optional): ambra api url. Defaults to "https://access.ambrahealth.com/api/v3".
        """
        self.sid = sid
        self.url = url
        self.log = []
        self.errors = []
        self.warnings = []
    def handle_ambra_request(self,endpoint:str,data:dict,throw=0,retry=0,last_error="",pg=1):
        """TODO: support pagation

        Args:
            endpoint (str): ambra api endpoint ex. /account/get
            data (dict): dictionary of parameters to pass to the endpoint
            throw (int, optional): flag to throw error. Defaults to 0.
            retry (int, optional): for recursion. Defaults to 0.
            last_error (str, optional): for recursion. Defaults to "".
            pg (int, optional): for recursion. Defaults to 1.

        Returns:
            dict: retuns the json response from the api if successful
        """
        
        if endpoint[0]!="/":
            endpoint = "/"+endpoint
        if isinstance(data,dict) and "sid" not in data:
            data["sid"] = self.sid
        # if retry>=3:
        #     if last_error not in self.errors:
        #         self.errors.append(last_error)
        #     if throw == 1:
        #         raise Exception(last_error)
        #     else:
        #         return
        try:
            response = requests.post(self.url+endpoint,data=data)
            if response.status_code == 401 and endpoint == '/account/get':
                raise Exception("sid error")
            res = response.json()
            res = sorted(res)
            if response.status_code == 200:
                if endpoint == "/bundle":
                    return res
                if res['status'] == "OK" and "list" in endpoint and "more" in res.keys():
                    pg+=1
                    data['page'] = pg 
                    return res.append(self.handle_ambra_request(endpoint,data,pg=pg))
                elif res['status']=="OK":
                    return res
                else:
                    raise Exception(str(response.status_code)+" - " + endpoint +" - " + res.get('status')+" - " + res.get('error_type') + "-"+str(res.get('error_subtype'))+"-"+json.dumps(data))
            elif res['error_type'] in ["INVALID_SETTING","INVALID_PERMISSION","INVALID_SETTING_VALUE"] and endpoint in ['/account/set','/role/set']:
                if res['error_type'] == "INVALID_SETTING":
                    settings = json.loads(data['settings'][0])
                    del settings[res['error_subtype']]
                    data['settings'] =[json.dumps(settings)]
                    # req[i]['URL'] = re
                elif res['error_type'] == "INVALID_PERMISSION":
                    data['permissions'] = json.loads(data['permissions'])
                    del data['permissions'][res['error_subtype']]
                    data['permissions'] =json.dumps(data['permissions'])
                elif res['error_type'] == "INVALID_SETTING_VALUE":
                    settings = json.loads(data['settings'][0])
                    del settings[res['error_subtype'].split(":")[0]]
                    data['settings'] =[json.dumps(settings)]    
                self.warnings.append(res.get('error_type')+ "- "+endpoint+" Automatic error resolution attempted, "+res.get('error_subtype')+" was removed from "+data['uuid']+" "+endpoint)
                self.handle_ambra_request(endpoint,data, retry=retry+1,throw = throw)
            elif res['error_type'] == "INVALID_CUSTOMFIELD" or (res['error_type'] == "INVALID_FIELD_NAME" and 'customfield' in str(res['error_subtype'])):
                # return first group match for (?:.*)(customfield-(?:\w|\d|\-)+) if not null
                if 'user' in endpoint:
                    id = data['user_id']
                else:
                    id = data['uuid']
                field_to_remove = re.match("(?:.*)(customfield-(?:\w|\d|\-)+)",str(res['error_subtype'])).group(1)
                if field_to_remove in data:
                    del data[field_to_remove]
                    
                    self.warnings.append(res.get('error_type')+ "- "+endpoint+" Automatic error resolution attempted, "+str(res['error_subtype'])+" was removed from "+id+" "+endpoint)
                    self.handle_ambra_request(endpoint,data, retry=retry+1,throw = throw)
                elif 'defaults' in data: 
                    defaults = json.loads(data['defaults'])
                    if field_to_remove in defaults:
                        del defaults[field_to_remove]
                        data['defaults'] = json.dumps(defaults)
                        self.warnings.append(res.get('error_type')+ "- "+endpoint+" Automatic error resolution attempted, "+str(res['error_subtype'])+" was removed from "+id+" "+endpoint)
                        self.handle_ambra_request(endpoint,data, retry=retry+1,throw = throw)
                else:
                    raise Exception(str(response.status_code)+" - " + endpoint +" - " + res.get('status')+" - " + res.get('error_type') + "-"+str(res.get('error_subtype'))+"-"+json.dumps(data))
            else:
                raise Exception(str(response.status_code)+" - " + endpoint +" - " + response.text + " - "+json.dumps(data))
        except Exception as err:
            
            if 'sid error' in err.args:
                print("SID ERROR")
                raise Exception("SID ERROR")
            if retry >= 2 and err not in self.errors:
                self.errors.append(err)
                if throw == 1:
                    raise Exception(last_error)
                return response.text
            output = self.handle_ambra_request(endpoint,data, retry=retry+1,throw = throw,last_error=err)
            return output
        
    # method for multiprocessing handle_ambra_request with input parameters requestsm endpoint, and threads
    def multiprocess_ambra_request(self,requests,endpoint,threads=30,pbar = False):
        if len(requests) < threads:
            threads = len(requests)
            if threads == 0:
                return []
        func = partial(self.handle_ambra_request,endpoint)
        p = ThreadPool(threads)
        if pbar == True:
            threaded_response = list(tqdm(p.imap(func,requests),total=len(requests)))
        else:
            threaded_response = list(p.map(func,requests))
        return threaded_response